#!/usr/bin/env python

#
# I simulate a Websensor EM01B.
#
# For usage, see my printUsage function or run me with my -h option.
#

import getopt
import asynchat
import asyncore
import os
import re
import signal
import socket
import sys

DEFAULT_PORT = 50437

def main(args):
  (port, isDebugOutputEnabled) = parseCommandLineArgs(args)
  d = ConnectionDispatcher(port, isDebugOutputEnabled)
  asyncore.loop()

def getProgramName(args=None):
  if args == None: args = sys.argv
  if len(args) == 0 or args[0] == "-c": return "PROGRAM_NAME"
  return os.path.basename(args[0])

def initializeSignalHandlers():
  signalToNameMap = {
    signal.SIGHUP:"SIGHUP",
    signal.SIGINT:"SIGINT",
    signal.SIGPIPE:"SIGPIPE",
    signal.SIGALRM:"SIGALRM",
    signal.SIGTERM:"SIGTERM",
    signal.SIGUSR1:"SIGUSR1",
    signal.SIGUSR2:"SIGUSR2"
  }

  def signalHandler(signum, frame):
    if signum in [signal.SIGINT, signal.SIGTERM]:
      print
      print "%s caught. Exiting." % signalToNameMap[signum]
      sys.exit(0)
    else:
      print >> sys.stderr
      print >> sys.stderr, "%s caught. Exiting." % signalToNameMap[signum]
      sys.exit(1)

  for each in signalToNameMap.keys():
    signal.signal(each, signalHandler)

def parseCommandLineArgs(args):
  (options, extra) = getopt.getopt(args[1:], "dp:h", ["debug", "port=",
    "help"])

  port = DEFAULT_PORT
  isDebugOutputEnabled = False

  for eachOptName, eachOptValue in options:
    if eachOptName in ("-d", "--debug"):
      isDebugOutputEnabled = True
    elif eachOptName in ("-p", "--port"):
      port = int(eachOptValue)
    elif eachOptName in ("-h", "--help"):
      printUsage(sys.stdout)
      sys.exit(0)

  if len(extra) > 0:
    print >> sys.stderr, "Error: unexpected command line argument \"%s\"" % \
      extra[0]
    printUsage(sys.stderr)
    sys.exit(1)

  return (port, isDebugOutputEnabled)

def printUsage(outStream):
  print >> outStream, """\
Usage: %s [-dph]

Options:
  -d,--debug        Print debug messages to stderr
  -p,--port=NUMBER  Listen on the specified port NUMBER for incoming
                    connections (default: %d)
  -h,--help         Print usage message and exit\
""" % (getProgramName(), DEFAULT_PORT)


class ConnectionDispatcher(asyncore.dispatcher):
  def __init__(self, port, debugOutputEnabled=False):
    asyncore.dispatcher.__init__(self)
    self.port = port
    self.debugOutputEnabled = debugOutputEnabled
    self.websensor = Websensor()
    self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
    self.set_reuse_addr()
    self.bind(("", port))
    self.listen(5)

  def handle_accept(self):
    ConnectionSession(self.accept(), self.websensor, self.debugOutputEnabled)


class ConnectionSession(asynchat.async_chat):
  SetTempRe = re.compile(r"Temp=(?P<val>[\d\.]+)$", re.I)
  SetTempUnitRe = re.compile(r"TempUnit=(?P<val>[FC])$", re.I)
  SetHumidRe = re.compile(r"Humid=(?P<val>[\d\.]+)$", re.I)
  SetIllumRe = re.compile(r"Illum=(?P<val>[\d\.]+)$", re.I)
  SetThermRe = re.compile(r"Therm=(?P<val>[\d\.]+)$", re.I)
  SetNoRspCntRe = re.compile(r"NoRspCnt=(?P<val>[\d]+)$", re.I)
  SetEmptyRspRe = re.compile(r"EmptyRsp=(?P<val>(?:yes)|(?:no))$", re.I)

  GetEmRe = re.compile(r"GET /index\.html\?em HTTP/1.[01]$", re.I)
  GetStatusJsonRe = re.compile(r"GET /status\.json HTTP/1.[01]$", re.I)
  GetStatusXmlRe = re.compile(r"GET /status\.xml HTTP/1.[01]$", re.I)
  RequestEndRe = re.compile(r"^$")

  def __init__(self, (conn, addr), websensor, debugOutputEnabled=False):
    asynchat.async_chat.__init__(self, conn)
    self.debugOutputEnabled = debugOutputEnabled
    self.set_terminator("\r\n")
    self.outputTerminator = "\r\n"
    self.websensor = websensor
    self.waitingForGetEmRequestEnd = False
    self.waitingForGetStatusJsonRequestEnd = False
    self.waitingForGetStatusXmlRequestEnd = False
    self.buffer = ""

  def collect_incoming_data(self, data):
    self.buffer = self.buffer + data

  def found_terminator(self):
    data = self.buffer
    self.buffer = ""
    if self.debugOutputEnabled:
      print >> sys.stderr, "< \"%s\"" % \
        (data + self.get_terminator()).encode("string_escape")
    self.handleClientRequest(data)

  def handleClientRequest(self, request):
    request = request.strip()

    match = self.SetTempRe.match(request)
    if match != None:
      self.websensor.setTemperature(float(match.group("val")))
      self.sendClientResponse("OK")
      return
    match = self.SetTempUnitRe.match(request)
    if match != None:
      self.websensor.setTemperatureUnit(match.group("val"))
      self.sendClientResponse("OK")
      return
    match = self.SetHumidRe.match(request)
    if match != None:
      self.websensor.setHumidity(float(match.group("val")))
      self.sendClientResponse("OK")
      return
    match = self.SetIllumRe.match(request)
    if match != None:
      self.websensor.setIllumination(float(match.group("val")))
      self.sendClientResponse("OK")
      return
    match = self.SetThermRe.match(request)
    if match != None:
      self.websensor.setThermistor(float(match.group("val")))
      self.sendClientResponse("OK")
      return
    match = self.SetNoRspCntRe.match(request)
    if match != None:
      self.websensor.setNoResponseCount(int(match.group("val")))
      self.sendClientResponse("OK")
      return
    match = self.SetEmptyRspRe.match(request)
    if match != None:
      self.websensor.setEmptyResponseEnabled(
        match.group("val").lower() == 'yes')
      self.sendClientResponse("OK")
      return

    match = self.GetEmRe.match(request)
    if match != None:
      self.waitingForGetEmRequestEnd = True
      return
    match = self.GetStatusJsonRe.match(request)
    if match != None:
      self.waitingForGetStatusJsonRequestEnd = True
      return
    match = self.GetStatusXmlRe.match(request)
    if match != None:
      self.waitingForGetStatusXmlRequestEnd = True
      return
    match = self.RequestEndRe.match(request)
    if match != None:
      if self.waitingForGetEmRequestEnd:
        if self.websensor.shouldRespond():
          if self.websensor.isEmptyResponseEnabled():
            self.sendClientResponse(
              "HTTP/1.0 200 OK\r\n" + \
              "Server: ES Web Server\r\n" + \
              "Content-Type: text/html\r\n" + \
              "Connection: close\r\n" + \
              "\r\n"
            )
          else:
            self.sendClientResponse((
              "HTTP/1.0 200 OK\r\n" + \
              "Content-Length: 124\r\n" + \
              "Server: ES Web Server\r\n" + \
              "Content-Type: text/html\r\n" + \
              "Connection: close\r\n" + \
              "\r\n" + \
              "<html>\r\n" + \
              "<head><title></title></head>\r\n" + \
              "<body>        EN1a\x03\xfd\x871T%s: %.1fHU:%.1f%%IL  %.1f" + \
                "                        </body></html>") % \
              (self.websensor.getTemperatureUnit(),
               self.websensor.getTemperature(),
               self.websensor.getHumidity(),
               self.websensor.getIllumination())
            )
        else:
          self.websensor.decrementNoResponseCount()
        self.close_when_done()
      elif self.waitingForGetStatusJsonRequestEnd:
        if self.websensor.shouldRespond():
          if self.websensor.isEmptyResponseEnabled():
            self.sendClientResponse(
              "HTTP/1.1 200 OK\n" + \
              "Access-Control-Allow-Origin: *\n" + \
              "Content-Type: application/json\n" + \
              "Connection: close\n" + \
              "\n"
            )
          else:
            tunVal = 0
            if self.websensor.getTemperatureUnit() == 'C': tunVal = 1
            self.sendClientResponse((
              "HTTP/1.1 200 OK\n" + \
              "Access-Control-Allow-Origin: *\n" + \
              "Content-Type: application/json\n" + \
              "Connection: close\n" + \
              "\n" + \
              "{" + \
                 "\"sid\":48190241," + \
                 "\"bid\":0," + \
                 "\"ver\":\"PR6X1.1.2\"," + \
                 "\"ssid\":\"\"," + \
                 "\"net\":0," + \
                 "\"sta\":0," + \
                 "\"wifi\":0," + \
                 "\"mfg\":0," + \
                 "\"shtSW\":0," + \
                 "\"s85SW\":0," + \
                 "\"si7SW\":1," + \
                 "\"thmSW\":1," + \
                 "\"ilmSW\":1," + \
                 "\"fldSW\":0," + \
                 "\"rtdSW\":0," + \
                 "\"mtrSW\":0," + \
                 "\"pirSW\":0," + \
                 "\"cinSW\":0," + \
                 "\"lvlSW\":0," + \
                 "\"vltSW\":0," + \
                 "\"vl3SW\":0," + \
                 "\"cvlSW\":0," + \
                 "\"somSW\":0," + \
                 "\"flwSW\":0," + \
                 "\"prsSW\":0," + \
                 "\"gasSW\":0," + \
                 "\"camSW\":0," + \
                 "\"bzrSW\":1," + \
                 "\"rlySW\":0," + \
                 "\"lcdSW\":1," + \
                 "\"tMod\":0," + \
                 "\"tmp\":%.2f," + \
                 "\"hum\":%.2f," + \
                 "\"thm\":%.2f," + \
                 "\"ilm\":%.2f," + \
                 "\"fld\":1," + \
                 "\"pir\":0," + \
                 "\"cin\":0," + \
                 "\"vlt\":1," + \
                 "\"vlt2\":0," + \
                 "\"vlt3\":0," + \
                 "\"cvl\":0.00," + \
                 "\"mtr\":0," + \
                 "\"flwType\":1," + \
                 "\"flw\":0.00," + \
                 "\"fVD\":0.00," + \
                 "\"fVT\":0.00," + \
                 "\"flvSW\":0," + \
                 "\"prsType\":1," + \
                 "\"prs\":0.00," + \
                 "\"gas\":0.00," + \
                 "\"gNm\":1," + \
                 "\"lvlType\":0," + \
                 "\"rtdType\":0," + \
                 "\"lvl\":0.00," + \
                 "\"rtd\":0.00," + \
                 "\"som\":0.00," + \
                 "\"tun\":%d," + \
                 "\"lun\":1," + \
                 "\"pun\":0," + \
                 "\"fun\":0," + \
                 "\"tmpAlm\":0," + \
                 "\"humAlm\":0," + \
                 "\"thmAlm\":0," + \
                 "\"ilmAlm\":0," + \
                 "\"fldAlm\":0," + \
                 "\"pirAlm\":0," + \
                 "\"cinAlm\":0," + \
                 "\"lvlAlm\":0," + \
                 "\"vltAlm\":0," + \
                 "\"vlt2Alm\":0," + \
                 "\"vlt3Alm\":0," + \
                 "\"flwAlm\":0," + \
                 "\"prsAlm\":0," + \
                 "\"gasAlm\":0," + \
                 "\"rtdAlm\":0," + \
                 "\"somAlm\":0," + \
                 "\"mtrAlm\":0," + \
                 "\"shtDet\":1," + \
                 "\"cinDet\":1," + \
                 "\"fldDet\":1," + \
                 "\"lvlDet\":1," + \
                 "\"thmDet\":1," + \
                 "\"flwDet\":0," + \
                 "\"prsDet\":0," + \
                 "\"mtrDet\":1," + \
                 "\"rawDat\":0.00" + \
              "}") % \
              (self.websensor.getTemperature(),
               self.websensor.getHumidity(),
               self.websensor.getThermistor(),
               self.websensor.getIllumination(),
               tunVal)
            )
        else:
          self.websensor.decrementNoResponseCount()
        self.close_when_done()
      elif self.waitingForGetStatusXmlRequestEnd:
        if self.websensor.shouldRespond():
          if self.websensor.isEmptyResponseEnabled():
            self.sendClientResponse(
              "HTTP/1.1 200 OK\n" + \
              "Access-Control-Allow-Origin: *\n" + \
              "Content-Type: application/xml\n" + \
              "Connection: close\n" + \
              "\n"
            )
          else:
            self.sendClientResponse((
              "HTTP/1.1 200 OK\n" + \
              "Access-Control-Allow-Origin: *\n" + \
              "Content-Type: application/xml\n" + \
              "Connection: close\n" + \
              "\n" + \
              "<?xml version=\"1.0\"?>" + \
              "<sensorsSW>" + \
                "<dvc>TBD</dvc>" + \
                "<sid0>48190241</sid0>" + \
                "<sht>inline</sht>" + \
                "<ilum>inline</ilum>" + \
                "<evin>none</evin>" + \
                "<evin3>none</evin3>" + \
                "<ethm>inline</ethm>" + \
                "<ecin>none</ecin>" + \
                "<efld>none</efld>" + \
                "<epir>none</epir>" + \
                "<stu0>OK</stu0>" + \
                "<tm0>%.2f</tm0>" + \
                "<hu0>%.2f</hu0>" + \
                "<il0>%.2f</il0>" + \
                "<tun0>%s</tun0>" + \
                "<cin>0</cin>" + \
                "<fin>1</fin>" + \
                "<pin>0</pin>" + \
                "<vin>0.00</vin>" + \
                "<thm>%.2f</thm>" + \
              "</sensorsSW>") % \
              (self.websensor.getTemperature(),
               self.websensor.getHumidity(),
               self.websensor.getIllumination(),
               self.websensor.getTemperatureUnit(),
               self.websensor.getThermistor())
            )
        else:
          self.websensor.decrementNoResponseCount()
        self.close_when_done()
      else:
        print >> sys.stderr, "error: got HTTP request end, but no GET before"
      return

    print >> sys.stderr, "error: unknown command \"%s\"" % \
      request.encode("string_escape")

  def sendClientResponse(self, response=""):
    data = response + self.outputTerminator
    self.push(data)
    if self.debugOutputEnabled:
      print >> sys.stderr, "> \"%s\"" % data.encode("string_escape")


class Websensor:
  def __init__(self):
    self.temperature = 81.0
    self.temperatureUnit = "F"
    self.humidity = 25.9
    self.illumination = 45.6
    self.thermistor = 75.0
    self.noResponseCount = 0
    self.emptyResponseEnabled = False

  def setTemperature(self, value):
    self.temperature = value

  def getTemperature(self):
    return self.temperature

  def setTemperatureUnit(self, value):
    self.temperatureUnit = value

  def getTemperatureUnit(self):
    return self.temperatureUnit

  def setHumidity(self, value):
    self.humidity = value

  def getHumidity(self):
    return self.humidity

  def setIllumination(self, value):
    self.illumination = value

  def getIllumination(self):
    return self.illumination

  def setThermistor(self, value):
    self.thermistor = value

  def getThermistor(self):
    return self.thermistor

  def setNoResponseCount(self, value):
    self.noResponseCount = value

  def getNoResponseCount(self):
    return self.noResponseCount

  def shouldRespond(self):
    return self.noResponseCount == 0

  def decrementNoResponseCount(self):
    if self.noResponseCount > 0:
      self.noResponseCount = self.noResponseCount - 1

  def setEmptyResponseEnabled(self, value):
    self.emptyResponseEnabled = value

  def isEmptyResponseEnabled(self):
    return self.emptyResponseEnabled


if __name__ == '__main__':
  try:
    initializeSignalHandlers()
    main(sys.argv)
  except Exception, e:
    if isinstance(e, SystemExit):
      raise e
    else:
      print >> sys.stderr
      print >> sys.stderr, "Error: %s" % e
      sys.exit(1)
